@@ -25,14 +25,67 @@ from networkx.algorithms.graph_hashing import weisfeiler_lehman_graph_hash
 import os
 import pandas as pd
 import numpy as np
+from splink_graph.utils import _laplacian_spectrum
 
 # setup to work around with pandas udf
 # see answers on
 # https://stackoverflow.com/questions/58458415/pandas-scalar-udf-failing-illegalargumentexception
 
 
+def cluster_basic_stats(df, src="src", dst="dst",cluster_id_colname="cluster_id",weight_colname="weight"):
+
+    """
+    
+input spark dataframe:
+
+
+|src|dst|cluster_id|
+|---|---|----------|
+|  e|  f|         0|
+
+    
+    
+output spark dataframe:
+
+|cluster_id|               nodes|nodecount|edgecount|density|
+|----------|--------------------|---------|---------|------|
+|8589934592|           [b, a, c]|        3|        2|0.666|
+|         0|[h, g, f, e, d, i..]|        7|        7|0.333|
+
+
+
+    
+    
+    """
+
+    edgec = df.groupby(cluster_id_colname).agg(f.count(weight_colname).alias("edgecount"))
+    srcdf = df.groupby(cluster_id_colname).agg(f.collect_set(src).alias("sources"))
+    dstdf = df.groupby(cluster_id_colname).agg(f.collect_set(dst).alias("destinations"))
+    allnodes = srcdf.join(dstdf, on=cluster_id_colname)
+    allnodes = allnodes.withColumn(
+        "nodes", f.array_union(f.col("sources"), f.col("destinations"))
+    ).withColumn("nodecount", f.size(f.col("nodes")))
+
+    output = allnodes.join(edgec, on=cluster_id_colname)
+
+    # density related calcs based on nodecount and max possible number of edges in an undirected graph
+
+    output = output.withColumn(
+        "maxNumberOfEdgesundir", f.col("nodecount") * (f.col("nodecount") - 1.0) / 2.0,
+    )
+    output = output.withColumn(
+        "density", f.round(f.col("edgecount") / f.col("maxNumberOfEdgesundir"), 3)
+    ).drop("sources", "destinations", "maxNumberOfEdgesundir")
+
+    return output
+
+
+
+
 def cluster_main_stats(sparkdf, src="src", dst="dst", cluster_id_colname="cluster_id"):
     """    
+    calculate diameter / transitivity(GCC) / triangle clustering coefficient LCC / square clustering coeff
+    and the weisfeiler-lehman graph hash of a cluster
 
     input spark dataframe:
 
@@ -116,13 +169,21 @@ def cluster_connectivity(
 ):
     """    
     node connectivity:
+    
     Measures the minimal number of vertices that can be removed to disconnect the graph.
     Larger vertex (node) connectivity --> harder to disconnect graph
     
     edge connectivity:
+    
     Measures the minimal number of edges that can be removed to disconnect the graph.
     Larger edge connectivity --> harder to disconnect graph
     
+    algebraic connectivity:
+    
+    The larger the algebraic connectivity, the more connected the graph is.
+    
+    efficiency:
+    
     The global efficiency of a graph is the average inverse distance between all pairs of nodes in the graph.
     The larger the average inverse shortest path distance, the more robust the graph.
     This can be viewed through the lens of network connectivity i.e., larger average inverse distance
@@ -160,6 +221,7 @@ def cluster_connectivity(
                 StructField("cluster_id", LongType()),
                 StructField("node_conn", IntegerType()),
                 StructField("edge_conn", IntegerType()),
+                StructField("algebraic_conn", FloatType()),
                 StructField("global_efficiency", FloatType()),
             ]
         ),
@@ -173,12 +235,15 @@ def cluster_connectivity(
         nc = nx.algorithms.node_connectivity(nxGraph)
         ec = nx.algorithms.edge_connectivity(nxGraph)
         ge = round(nx.global_efficiency(nxGraph), 3)
-
+        
+        lapl_spc = _laplacian_spectrum(nxGraph) 
+        ac = round(lapl_spc[1], 3) # calculate algebraic connectivity
+        
         co = pdf[cluster_id_colname].iloc[0]  # access component id
 
         return pd.DataFrame(
-            [[co] + [nc] + [ec] + [ge]],
-            columns=["cluster_id", "node_conn", "edge_conn", "global_efficiency"],
+            [[co] + [nc] + [ec] + [ac] + [ge]],
+            columns=["cluster_id", "node_conn", "edge_conn","algebraic_conn", "global_efficiency"],
         )
 
     out = sparkdf.groupby(cluster_id_colname).apply(conn_eff)
@@ -229,8 +294,19 @@ def cluster_modularity(
         nxGraph = nx.Graph()
         nxGraph = nx.from_pandas_edgelist(pdf, psrc, pdst, pdistance)
 
-        ## TODO: comments! to document this code
-
+        ## TODO: document this code
+        # this is a method that calculates the modularity of a cluster if partitioned into 2 parts 
+        # where the split is happening where the highest edge betweeness is.
+        
+        # if modularity is negative :
+        #      that means that the split just leaves singleton nodes or something like that. 
+        #      basically the cluster is of no interest
+        # if modularity is 0 or very close to 0 :
+        #      its a cluster of well connected nodes so... nothing to see here really.
+        # if modularity is around 0.4+ then :
+        #      its a cluster of possible interest
+        
+        
         def most_central_edge(G):
             centrality = edge_betweenness_centrality(
                 G, weight=pdistance, normalized=True
